//****************************************************************************
// File: diaolg.c
//
// Purpose: Both parent and child dialog window's message handler
//
// Functions:
//    ParentDlgProc() - Parent dialogs window procedure
//    ChildDlgProc() - Child dialogs window procedure
//    ButtonSubProc() - Last button on the Parent dialogs subclass procedure
//    wDoDialog() - called by DoCommands to create and display the dialog boxes
//
// Development Team:
//    Homi Hanjani-Rassouli.
//
// Written by Microsoft Product Support Services, Windows Developer Support
// Copyright (c) 1992 Microsoft Corporation. All rights reserved.
//****************************************************************************

#include <windows.h>
#include "dlgtab.h"
#include "global.h"
#include "dialog.h"

//****************************************************************************
// Function: ParentDlgProc
//
// Purpose: Dialog procedure that subclasses the Button control to
//          take care of tabbing problem.
//
// Parameters:
//    hDlg     == Handle to _this_ window (Parent dialog box).
//    imessage == Message to process.
//    wParam   == WORD parameter -- depends on message
//    lParam   == LONG parameter -- depends on message
//
// Returns:
//  Standard for dialog procedure.
//
// Comments:
//
// History:  Date       Author        Reason
//           11/09/92   HHR           Created
//****************************************************************************

BOOL FAR PASCAL ParentDlgProc(HWND hDlg, UINT iMessage,
                              WPARAM wParam, LPARAM lParam)
    {
    switch (iMessage)
        {
        case WM_INITDIALOG:
            CheckRadioButton(hDlg, ID_METHODDEF, ID_METHODSUB, ID_METHODSUB);
            return(TRUE);

        case WM_COMMAND:
            switch(wParam)
                {
                case ID_METHODSUB:
                    //Hook in the subclass procedure for the Cancel Button.
                    SetWindowLong(GetDlgItem(hDlg,ID_CHILDDLG),
                                  GWL_WNDPROC, (LONG)lpButtonSubProc);

                    return(TRUE);

                    break;

                case ID_METHODDEF:
                    //Unhook the subclass procedure for the Cancel Button.
                    SetWindowLong(GetDlgItem(hDlg,ID_CHILDDLG),
                                  GWL_WNDPROC, (LONG)lpButtonProcOrg);

                    return(TRUE);

                    break;

                case ID_BUTTON1:
                case ID_BUTTON2:
                case ID_BUTTON3:
                case ID_EDIT:
                    break;

                case ID_CHILDDLG:
                    iDoDialog(CHILDDLG);
                    break;

                case IDOK:
                    //Unhook the subclass procedure for the Cancel Button
                    //before exiting.
                    SetWindowLong(GetDlgItem(hDlg, ID_CHILDDLG),
                                  GWL_WNDPROC, (LONG)lpButtonProcOrg);
                    FreeProcInstance(lpButtonSubProc);

                    //Close the child dialog if it is stil open and reset
                    //the handle to NULL.
                    if (hChildDlg)
                      SendMessage(hChildDlg,WM_CLOSE, 0, 0L);
                    DestroyWindow(hDlg);
                    hParentDlg = NULL;

                    return TRUE;

                default:
                    break;
                }
            break;

        default:
            break;
        }

    return FALSE;
    }

//****************************************************************************
// Function: ChildDlgProc
//
// Purpose:  Processes messages for Child dialog box
//
// Parameters:
//    hWnd    == Handle to _this_ window (Child dialog procedure).
//    message == Message to process.
//    wParam  == WORD parameter -- depends on message
//    lParam  == LONG parameter -- depends on message
//
// Messages:
//     WM_INITDIALOG - initialize dialog box
//     WM_COMMAND    - Input received
//     WM_CLOSE      - Closing the dialog box and clean up
//
// Returns:
//  Standard for dialog procedure.
//
// Comments:
//	 No initialization is needed for this particular dialog box, but TRUE
//	 must be returned to Windows.
//
//	 Wait for user to click on "Ok" button, then close the dialog box.
//
// History:  Date       Author        Reason
//           11/09/92   HHR           Created
//****************************************************************************

BOOL FAR PASCAL ChildDlgProc(hDlg, message, wParam, lParam)
HWND hDlg;                               // window handle of the dialog box 
UINT message;                        // type of message                 
WPARAM wParam;                             // message-specific information    
LPARAM lParam;
{
   switch (message) {
      case WM_INITDIALOG:                // message: initialize dialog box
         return (TRUE);

      case WM_COMMAND:                   // message: received a command
         if ((wParam == IDOK) ||         // "OK" box selected?
             (wParam == IDCANCEL))       // System menu close command?
            {
            DestroyWindow(hDlg);         // Exits the dialog box
            hChildDlg = NULL;            // Reset the handle to NULL
            return (TRUE);
            }

         break;

      case WM_CLOSE:                     // message: recieved a close
         DestroyWindow(hDlg);
         hChildDlg = NULL;
         return (TRUE);
        
   }

   return (FALSE);                       // Didn't process a message
}

//****************************************************************************
// Function: ButtonSubProc
//
// Purpose: Dialog procedure that subclasses the Cancel Button to detect the
// TAB and Arrow keys..
//
// Parameters:
//  Standard for a window procedure.
//
// Returns:
//  Standard for dialog procedure.
//
// Comments:
//
// History:  Date       Author        Reason
//           11/09/92   HHR           Created
//****************************************************************************

LONG FAR PASCAL ButtonSubProc(HWND hWnd, WORD iMessage,
                                WORD wParam, LONG lParam )
{
   LRESULT lRet;
   WORD    wRet;
   LPMSG   lpmsg;
   HWND    hwndCtrls;
   HWND    hwndBtn1;

   switch (iMessage)
      {
      case WM_GETDLGCODE:
         lRet = CallWindowProc(lpButtonProcOrg,hWnd,iMessage,wParam,lParam);
         
         if (lParam)
            {
            lpmsg= (LPMSG) lParam;
            if (lpmsg->message == WM_KEYDOWN)
               {
               if ((lpmsg->wParam == VK_TAB)   ||
                   (lpmsg->wParam == VK_DOWN)  || (lpmsg->wParam == VK_UP) ||
                   (lpmsg->wParam == VK_RIGHT) || (lpmsg->wParam == VK_LEFT) )
                  lRet |= DLGC_WANTMESSAGE;
               }
            }
         return (lRet);

      case WM_KEYDOWN:

         // if it the arrow keys are used then just beep and do not
         // change the focus since this is how it would have worked
         // if the next control had WS_GROUP.
         if ( (wParam == VK_DOWN)  || (wParam == VK_UP) ||
              (wParam == VK_RIGHT) || (wParam == VK_LEFT) )
            {
            MessageBeep(0);
            return(0L);
            }
 
         // if it is the TAB key then change the focus to the first
         // control.
         if ( (wParam == VK_TAB) )
           {
            // Set the focus to the first control
            hwndBtn1 = GetDlgItem(hParentDlg,ID_BUTTON1);
            
            SetFocus(hwndBtn1);

            // Loop through all the controls and remove the dark border 
            // that the previous default push button has.
            hwndCtrls = GetWindow(hParentDlg, GW_CHILD);
            while (hwndCtrls)
                {
                wRet = (WORD)(DWORD)SendMessage(hwndCtrls, WM_GETDLGCODE,
                                                0, 0L);
 
               if (wRet & DLGC_DEFPUSHBUTTON)
                 SendMessage(hwndCtrls, BM_SETSTYLE,  (WPARAM)BS_PUSHBUTTON,
                             TRUE);
                          
               hwndCtrls = GetWindow(hwndCtrls, GW_HWNDNEXT);
               }
 
           // Give the hwndBtn1 button the dault push button border
            SendMessage(hwndBtn1,  BM_SETSTYLE,(WPARAM)BS_DEFPUSHBUTTON,TRUE);
            return(0L);
            }

         break;
        
      default:
         break;
      }
 
   return CallWindowProc(lpButtonProcOrg,hWnd,iMessage,wParam,lParam);
}

//****************************************************************************
// Function:  iDoDialog
//
// Purpose: Handles necessary housekeeping to display a dialog, i.e. making
//          the ProcInstance and calling DialogBox.  Also processes the
//          return value in case of errors.
//
// Parameters: 
//  iWhichDlg  Determines which dialog should be created
//
// Returns:
//  -1      If the function fails to display the dialog.
//          Otherwise the value that the dialog passes from EndDialog.
//
// Comments:
//
// History:  Date       Author        Reason
//           11/09/92   HHR           Created
//****************************************************************************

int PASCAL iDoDialog(int iWhichDlg)
{
   int        iRet = -1;

   switch (iWhichDlg)
      {
      case PARENTDLG:
         {
         LONG       lOld;

         hParentDlg=CreateDialog(ghInst, MAKEINTRESOURCE(IDD_PARENTDLG), ghWnd, lpParentDlgProc);
         if (hParentDlg)
            {
            iRet = 1;

            // Subclassing the last button in the parent dialog.
            lpButtonSubProc = MakeProcInstance((FARPROC)ButtonSubProc,ghInst);

            // Get the handle to the Button's window procedure.
            lOld=GetWindowLong(GetDlgItem(hParentDlg, ID_CHILDDLG), GWL_WNDPROC);
            lpButtonProcOrg=(FARPROC) lOld;

            //Hook in the subclass procedure for the Cancel Button.
            SetWindowLong(GetDlgItem(hParentDlg,ID_CHILDDLG), GWL_WNDPROC,
                          (LONG)lpButtonSubProc);
            }

         break;
         }

      case CHILDDLG:
         // if the parent dialog exists and no child dialog has been created
         // yet create the child dialog.
         if ((hParentDlg) && (!hChildDlg))
            {

            // creating the Child dialog box.
    
            hChildDlg = CreateDialog(ghInst,                  // current instance
                                MAKEINTRESOURCE(IDD_CHILDDLG),// resource to use
                                hParentDlg,                   // parent handle
                                lpChildDlgProc);              // instance address
    
            if (hChildDlg)
               {
               RECT rcParent;
               RECT rcChild;
      
               iRet = 1;
               // set the parent of the child dialog.
               SetParent(hChildDlg,hParentDlg);
      
               // Get the client area of the parent dialog and the size of
               // the child dialog.
               GetClientRect(hParentDlg,&rcParent);
               GetWindowRect(hChildDlg,&rcChild);
               ScreenToClient(hParentDlg,(LPPOINT)&rcChild.left);
               ScreenToClient(hParentDlg,(LPPOINT)&rcChild.right);
      
               // Then position the child in the right hand corner of the 
               // parent dialog, give it the focus and show it.
               MoveWindow(hChildDlg,
                          rcParent.left + 300,          // new left pos of child dialog
                          rcParent.top  +   5,          // new top pos of child dialog
                          rcChild.right - rcChild.left, // new width of child dialog
                          rcChild.bottom - rcChild.top, // new height of child dialog
                          TRUE);
               SetFocus(hParentDlg);
               ShowWindow(hChildDlg,SW_SHOW);
               }
            }
         break;

      }  // Just finished attempting to Create a dialog.
    
   return iRet;
}


