//***************************************************************************
//
// FILE NAME: DLGPROCS.C
//
//    DLGPROCS.C holds the dialog procedure for the main dialog, and the
//    "fake" dialog procedures for the child dialogs.  It includes the
//    following functions:
//
//
// FUNCTIONS:
//
//    MainDlgProc ()     The main dialog's procedure.  This is a true dialog
//                       proc.
//
//    IdentityDlgProc () The "identity" child dialog's procedure.  This may
//                       be treated like a normal Windows callback procedure
//                       in most respects, but it is actually called from
//                       MainDlgProc().
//
//    PrefsDlgProc ()    The "preferences" child dialog's procedure.  See
//                       above.
//
//    WishesDlgProc ()   The "wishes" child dialog's procedure.  See above.
//
//    About ()           The About Box callback for this sample.
//
// AUTHOR: John Dempsey
//
// Written by Microsoft Product Support Services, Windows Developer Support.
// Copyright (c) 1993 Microsoft Corporation.  All rights reserved.
//
//***************************************************************************

#include <windows.h>
#include "multidlg.h"
#include "dlgprocs.h"

extern DLGINFO diDlgs[] ;

//***************************************************************************
//
// FUNCTION: MainDlgProc()
//
// PURPOSE:
//     Is the Windows callback function for the main dialog.  Handles the
//     initialization of the listbox, and handles the task of making child
//     dialogs visible and invisible as appropriate.  This function also
//     passes any unused messages on to the current active child dialog.  In
//     In this way code supporting different child dialogs can be neatly
//     organized.
//
// PARAMETERS:
//     Standard Windows callback parameters.
//
// RETURN VALUE:
//     Standard dialog return value: TRUE if we handled the message and don't
//     want DefDlgProc() to get it, FALSE if we want DefDlgProc() to handle
//     it.
//
// COMMENTS:
//     Note that we return 0 on WM_INITDIALOG to prevent Windows from
//     setting the focus.  Also note that we return to Windows whatever our
//     child dialog procedure returns to us in cases where we passed the
//     message along to it.
//
// HISTORY: Date    Author  Reason
//          6/1/93  JRD     Created
//
//***************************************************************************

BOOL __export CALLBACK MainDlgProc (HWND hDlg,
                                    UINT message,
                                    WPARAM wParam,
                                    LPARAM lParam)
{
    static HWND    hListbox ;
           LRESULT lCurSel ;
           int     i = 0 ;
    
    switch (message)
        {
        case WM_INITDIALOG:
            hListbox = GetDlgItem (hDlg, IDC_LISTBOX) ;
            while (diDlgs[i].iResID)
                {
                SendMessage (hListbox, LB_ADDSTRING, 0,
                             (LPARAM) (LPSTR) diDlgs[i].szListText) ;
                i++ ;
                }

            SendMessage (hListbox, LB_SETCURSEL, 0, 0L) ;
            
            // Selections made with LB_SETCURSEL do not generate a
            // LBN_SELCHANGE message.  Generate one manually.
            
            PostMessage (hDlg,
                         WM_COMMAND,
                         IDC_LISTBOX,
                         MAKELONG(hListbox, LBN_SELCHANGE)) ;
                         
            // Focus needs to be set back to the listbox after the creation
            // of the first child dialog is complete.  Technically, it's
            // there, but it doesn't LOOK there.  So we do this.

            PostMessage (hDlg, WM_MYFOCUS, 0, 0L) ;                         
                         
            return 0 ; // I set the focus myself (or at least I will).
            
         case WM_MYFOCUS:
            SetFocus (hListbox) ;
            return TRUE ; 
                        
        case WM_COMMAND:           
            switch (wParam)
                {
                case IDOK:
                    // Save values and fallthru...
                    
                case IDCANCEL:
                    // Free the handle/text lists and fPresent flags (the
                    // controls themselves will die automatically as children
                    // of the dialog).
                    
                    while (diDlgs[i].iResID)
                        {
                        if (diDlgs[i].fPresent)
                            {
                            LocalFree ((HLOCAL) diDlgs[i].pCtrlHndls) ;
                            LocalFree ((HLOCAL) diDlgs[i].pCtrlTexts) ;
                            diDlgs[i].pCtrlHndls = NULL ;
                            diDlgs[i].pCtrlTexts = NULL ;
                            diDlgs[i].fPresent = FALSE ;
                            }
                        i++ ;
                        }
                    iCurDlg = -1 ;
                                            
                    EndDialog (hDlg, 0) ;
                    return TRUE ;
                    
                case IDC_LISTBOX:
                    if (LBN_SELCHANGE == HIWORD (lParam))
                        {
                        lCurSel = SendMessage (hListbox, LB_GETCURSEL, 0, 0L) ;

                        if (-1 == lCurSel)
                            return FALSE ;

                        // lCurSel++ ;

                        if (lCurSel != iCurDlg)
                            {
                            ChildDialogVisible (FALSE, hDlg, iCurDlg) ;
                            iCurDlg = (int) lCurSel ;
                            ChildDialogVisible (TRUE, hDlg, iCurDlg) ;
                            }
                        return TRUE ;
                        }
                }
        }

    // Anything not handled by the main dialog is passed to the appropriate
    // child dialog procedure.
    if (iCurDlg > 0)
        return ( (*(diDlgs[iCurDlg].fpDlgProc))(hDlg, message, wParam, lParam) ) ;
    else
        return 0L ;     
}


//***************************************************************************
//
// FUNCTION: IdentityDlgProc()
//
// PURPOSE:
//     Handles messages specific to the Identity child dialog.
//
// PARAMETERS:
//     Same as standard Windows callback function.
//
// RETURN VALUE:
//     Standard dialog return value.
//
// COMMENTS:
//     This is actually called by MainDlgProc().
//
// HISTORY: Date    Author  Reason
//          6/1/93  JRD     Created
//
//***************************************************************************

BOOL IdentityDlgProc (HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    return FALSE ;
}    


//***************************************************************************
//
// FUNCTION: PrefsDlgProc()
//
// PURPOSE:
//     Handles messages specific to the Preferences child dialog.
//
// PARAMETERS:
//     Same as standard Windows callback function.
//
// RETURN VALUE:
//     Standard dialog return value.
//
// COMMENTS:
//     This is actually called by MainDlgProc().
//
// HISTORY: Date    Author  Reason
//          6/1/93  JRD     Created
//
//***************************************************************************

BOOL PrefsDlgProc (HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
        {
        // I get faked WM_INITDIALOGs when the children are created.  I
        // can use them for initializing my controls.

        case WM_INITDIALOG:
            CheckRadioButton (hDlg,
                              IDC_P_SCRAMBLED,
                              IDC_P_RAW,
                              IDC_P_SCRAMBLED) ;
                              
            CheckRadioButton (hDlg,
                              IDC_P_PAPER,
                              IDC_P_CLOTH,
                              IDC_P_CLOTH) ;
            return TRUE ;                             

        case WM_COMMAND:
            if (IDC_P_DINGS == wParam)
                if (BN_CLICKED == HIWORD(lParam))
                    {
                    MessageBeep (0) ;
                    return TRUE ;
                    }
        }

    return FALSE ;
}


//***************************************************************************
//
// FUNCTION: WishesDlgProc()
//
// PURPOSE:
//     Handles messages specific to the Wishes child dialog.
//
// PARAMETERS:
//     Same as standard Windows callback function.
//
// RETURN VALUE:
//     Standard dialog return value.
//
// COMMENTS:
//     This is actually called by MainDlgProc().
//
// HISTORY: Date    Author  Reason
//          6/1/93  JRD     Created
//
//***************************************************************************
  
BOOL WishesDlgProc (HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
    if (WM_INITDIALOG == message)
        return TRUE ;

    return FALSE ;
}


//***************************************************************************
//
// FUNCTION: About()
//
// COMMENTS:
//     This is a standard Windows dialog procedure.
//
// HISTORY: Date    Author  Reason
//          6/1/93  JRD     Created
//
//***************************************************************************

BOOL __export CALLBACK About (HWND hDlg,
                              UINT message,
                              WPARAM wParam,
                              LPARAM lParam)
{
    switch (message)
        {
        case WM_INITDIALOG:
            return TRUE ;
            
        case WM_COMMAND:           
            switch (wParam)
                {
                case IDOK:
                case IDCANCEL:
                    EndDialog (hDlg, 0) ;
                }
        }

    return FALSE ;
}

