/*  SuperCls.c
 *  SuperClass Application
 *  Windows Toolkit Version 3.00
 *  Copyright (c) Microsoft 1985-1991. All rights reserved.
 *
 */

/*
   This sample application demonstrates the idea of "superclassing."
   Superclassing provides the same capabilities as does subclassing, but
   can be much more convenient to use and produce smaller code if there
   are multiple controls to be altered in the same fashion.

   To superclass, one creates a new class which uses the control window
   procedure instead of the default window procedure. This enables the
   application to process the appropriate messages and to pass the others
   along to the default procedure where the messages will be processed
   correctly based on the control type.

   To implement superclassing, create a window of the type to superclass.
   Query all available information about the control and use this
   information when the new window is created, substituting the class
   name and window procedure. When this window procedure is called,
   process the desired messages and pass the others to the standard
   control procedure (determined in the query phase).

   By registering a new class, the application can create superclassed
   controls using the application resource (RC) file or CreateWindow()
   function without any further work which can eliminate a number of
   SetWindowLong() calls if many controls are to be modified in the
   same manner.

   In this example, the application creates a new control class called
   "SuperEdit" which is identical to a regular edit control but refuses
   to accept the "b" character.

   Look for the string "SuperEdit" in the .C, .RC, and .DEF files to see
   the affected areas of the code.

   This example shows superclassed Windows controls. Because the class
   name has been changed and the application's instance is used in the
   class registration, all controls will be created with this application's
   DS. This causes no problems for controls. However, this does not work
   if an attempt is made to superclass something other than a control.
   In that case, it is desirable to use the application's instance with
   the class so that it will be destroyed when the application exits.
   However, it is necessary to perform some additional work before the
   call to CallWindowProc() to set the DS for the subclassed window
   procedure. The hInstance passed to CreateWindow() determines which DS
   the superclassing procedure uses; to have the superclassed procedure
   use a different DS, it is necessary to change the DS value. For more
   information, see the notes near the call to CallWindowProc().

   Each edit control created using DialogBox(), CreateDialog(), or through
   an RC file normally has its own DS. When the application changes the
   class name, the code within Windows which handles edit controls is not
   activated. Instead, the control uses the application's DS as if it
   were created using CreateWindow(). This does not cause any difficulties
   except that when the code for the superclassed edit control calls
   LocalAlloc(), it consumes space in the application's local heap.
   This should not cause problems in most cases.
 */
#include "windows.h"
#include "supercls.h"

char szAppName[32];
char szAbout[32];
char szMessage[32];
int MessageLength;
HWND hWndMain;
static HANDLE hInst;
FARPROC lpprocAbout;
FARPROC lpprocDialog;

long FAR PASCAL SuperClsWndProc(HWND, unsigned, WORD, LONG);
long FAR PASCAL SuperEditWndProc(HWND, unsigned, WORD, LONG);
BOOL InitApp(HANDLE);
BOOL InitSuperCls(HANDLE);
void SuperClsPaint(HDC);
FARPROC lpEditWndProc;


BOOL FAR PASCAL About(hDlg, message, wParam, lParam)
HWND hDlg;
unsigned message;
WORD wParam;
LONG lParam;
{
   if (message == WM_COMMAND)
   {
      EndDialog(hDlg, TRUE);
      return TRUE;
   }
   else if (message == WM_INITDIALOG)
      return TRUE;
   else
      return FALSE;
}


BOOL FAR PASCAL Dialog(hDlg, message, wParam, lParam)
HWND hDlg;
unsigned message;
WORD wParam;
LONG lParam;
{
   switch (message)
      {
   case WM_INITDIALOG:
      break;

   case WM_COMMAND:
      switch (wParam)
         {
      case IDOK:
         EndDialog(hDlg, TRUE);
         break;

      case IDCANCEL:
         EndDialog(hDlg, FALSE);
         break;

      default:
         return FALSE;
         }
      break;

   default:
      return FALSE;
      }
   return TRUE;
}


void SuperClsPaint(hDC)
HDC hDC;
{
   TextOut(hDC, (short)10, (short)10, (LPSTR)szMessage, (short)MessageLength);
}


BOOL InitSuperCls(hInstance)
HANDLE hInstance;
{
   HWND hEditWnd;
   WNDCLASS Class;

   /* Superclassing done here... */

   /* Create this window to get information about it. */

   hEditWnd = CreateWindow((LPSTR)"EDIT", (LPSTR)" ", WS_CHILD,
                           0, 0, 0, 0,
                           (HWND)hWndMain,    /* no parent */
                           0,                 /* use class menu */
                           (HANDLE)hInstance, /* handle to window instance */
                           (LPSTR)NULL);      /* no params to pass on */

   /* Get the information necessary from the class, and create new class. */

   /* Save this for use later in the new window procedure */
   lpEditWndProc = (FARPROC)GetWindowLong(hEditWnd, GWL_WNDPROC);

   /*
    * Register the new superclass of the standard edit control.
    * It is the same as the original edit control except for the
    * window procedure, hInstance and class name.
    *
    * It is necessary to be very careful about the window extra bytes.
    * If the new superclassing procedure does not use them, the code
    * below is correct. If the new superclass procedure requires extra
    * bytes, it must note how many bytes the standard control procedure
    * uses and add the additional bytes to that number. The code must
    * refer to these bytes as an offset from the number of bytes the
    * standard control uses (for example cbClsExtraEdit+0, cbClsExtraEdit+2,
    * and so forth).
    */
   Class.style = GetClassWord(hEditWnd, GCW_STYLE);
   Class.lpfnWndProc = SuperEditWndProc;
   Class.cbClsExtra = GetClassWord(hEditWnd, GCW_CBCLSEXTRA);
   Class.cbWndExtra = GetClassWord(hEditWnd, GCW_CBWNDEXTRA);

   /*
    * Use this application's instance to have the window procedure
    * use the application's DS. Also, the class will be destroyed when
    * the last instance exits.
    *
    * Note: Classes are destroyed at application exit by searching the
    * list of classes for any classes with the current hModule (derived
    * from hInstance). Hence, to destroy the "SuperEdit" class, set
    * the class hInstance to the application's hInstance. The internal
    * class structure actually stores hModule rather than hInstance.
    * (RegisterClass() converts hInstance to hModule.
    */
   Class.hInstance = hInstance;
   Class.hIcon = GetClassWord(hEditWnd, GCW_HICON);
   Class.hCursor = GetClassWord(hEditWnd, GCW_HCURSOR);
   Class.hbrBackground = GetClassWord(hEditWnd, GCW_HBRBACKGROUND);
   Class.lpszMenuName = NULL; /* It doesn't have one. */
   Class.lpszClassName = (LPSTR)"SuperEdit";

   /* Window not needed any longer. */
   DestroyWindow(hEditWnd);

   if (!RegisterClass((LPWNDCLASS)&Class))
   {
      /* Initialization failed.
       * Windows will automatically deallocate all allocated memory.
       */
      return FALSE;
   }
   return TRUE;
}


/* Procedure called when the application is loaded for the first time */
BOOL InitApp(hInstance)
HANDLE hInstance;
{
   WNDCLASS Class;

   /* Load strings from resource */
   LoadString(hInstance, IDSNAME, (LPSTR)szAppName, sizeof(szAppName));
   LoadString(hInstance, IDSABOUT, (LPSTR)szAbout, sizeof(szAbout));
   MessageLength = LoadString(hInstance, IDSTITLE, (LPSTR)szMessage,
                              sizeof(szMessage));
   Class.hCursor = LoadCursor(NULL, IDC_ARROW);
   Class.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(SUPERCLSICON));
   Class.lpszMenuName = (LPSTR)szAppName;
   Class.lpszClassName = (LPSTR)szAppName;
   Class.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
   Class.hInstance = hInstance;
   Class.style = CS_HREDRAW | CS_VREDRAW;
   Class.cbClsExtra = 0;
   Class.cbWndExtra = 0;
   Class.lpfnWndProc = SuperClsWndProc;
   if (!RegisterClass((LPWNDCLASS)&Class))
   {
      /* Initialization failed.
       * Windows will automatically deallocate all allocated memory.
       */
      return FALSE;
   }
   return TRUE;        /* Initialization succeeded */
}


int PASCAL WinMain(hInstance, hPrevInstance, lpszCmdLine, cmdShow)
HANDLE hInstance, hPrevInstance;
LPSTR lpszCmdLine;
int cmdShow;
{
   MSG msg;
   HWND hWnd;
   HMENU hMenu;

   if (!hPrevInstance)
   {
      /* Call initialization procedure if this is the first instance */
      if (!InitApp(hInstance))
         return FALSE;
   }
   else
   {
      /* Copy data from previous instance */
      GetInstanceData(hPrevInstance, (PSTR)szAppName, sizeof(szAppName));
      GetInstanceData(hPrevInstance, (PSTR)szAbout, sizeof(szAbout));
      GetInstanceData(hPrevInstance, (PSTR)szMessage, sizeof(szMessage));
      GetInstanceData(hPrevInstance, (PSTR)&MessageLength,
                      sizeof(MessageLength));
   }
   hWnd = CreateWindow((LPSTR)szAppName, (LPSTR)szMessage,
                       WS_OVERLAPPEDWINDOW,
                       CW_USEDEFAULT, CW_USEDEFAULT,
                       CW_USEDEFAULT, CW_USEDEFAULT,
                       (HWND)NULL,        /* no parent */
                       (HMENU)NULL,       /* use class menu */
                       (HANDLE)hInstance, /* handle to window instance */
                       (LPSTR)NULL);      /* no params to pass on */

   /* Save instance handle for DialogBox */
   hInst = hInstance;
   hWndMain = hWnd;

   /* prepare for superclassing */
   if (!hPrevInstance)
   {
      /* Call initialization procedure if this is the first instance */
      if (!InitSuperCls(hInstance))
         return FALSE;
   }
   else
   {
      /* Copy data from previous instance */
      GetInstanceData(hPrevInstance, (PSTR)&lpEditWndProc,
                      sizeof(lpEditWndProc));
   }

   /* Bind callback function with module instance */
   lpprocAbout = MakeProcInstance((FARPROC)About, hInstance);
   lpprocDialog = MakeProcInstance((FARPROC)Dialog, hInstance);

   /* Insert "About..." into system menu */
   hMenu = GetSystemMenu(hWnd, FALSE);
   ChangeMenu(hMenu, 0, NULL, 999, MF_APPEND | MF_SEPARATOR);
   ChangeMenu(hMenu, 0, (LPSTR)szAbout, IDSABOUT, MF_APPEND | MF_STRING);

   /* Make window visible according to the way the app is activated */
   ShowWindow(hWnd, cmdShow);
   UpdateWindow(hWnd);

   /* Polling messages from event queue */
   while (GetMessage((LPMSG)&msg, NULL, 0, 0))
   {
      TranslateMessage((LPMSG)&msg);
      DispatchMessage((LPMSG)&msg);
   }
   FreeProcInstance(lpprocAbout);
   FreeProcInstance(lpprocDialog);
   return (int)msg.wParam;
}


/* Procedures which make up the window class. */

long FAR PASCAL SuperClsWndProc(hWnd, message, wParam, lParam)
HWND hWnd;
unsigned message;
WORD wParam;
LONG lParam;
{
   PAINTSTRUCT ps;

   switch (message)
      {
   case WM_SYSCOMMAND:
      switch (wParam)
         {
      case IDSABOUT:
         DialogBox(hInst, MAKEINTRESOURCE(ABOUTBOX), hWnd, lpprocAbout);
         break;

      default:
         return DefWindowProc(hWnd, message, wParam, lParam);
         }
      break;

   case WM_COMMAND:
      switch (wParam)
         {
      case IDMDIALOG:
         DialogBox(hInst, MAKEINTRESOURCE(IDDDIALOG), hWnd, lpprocDialog);
         break;

      default:
         return DefWindowProc(hWnd, message, wParam, lParam);
         }
      break;

   case WM_DESTROY:
      PostQuitMessage(0);
      break;

   case WM_PAINT:
      BeginPaint(hWnd, (LPPAINTSTRUCT)&ps);
      SuperClsPaint(ps.hdc);
      EndPaint(hWnd, (LPPAINTSTRUCT)&ps);
      break;

   default:
      return DefWindowProc(hWnd, message, wParam, lParam);
      break;
      }
   return (0L);
}


/* The superclassing procedure. */

long FAR PASCAL SuperEditWndProc(hWnd, message, wParam, lParam)
HWND hWnd;
unsigned message;
WORD wParam;
LONG lParam;
{
   long Result = 0L;

   switch (message)
      {
   /*
    * At this point, it is possible to trap any message desired which
    * is the same as subclassing. As an example, make the edit controls
    * refuse "b" characters. 
    */
   case WM_CHAR:
      if (wParam == 'b')
      {
         MessageBeep(1);
         break;
      }

   /* Fall through for messages not handled explicitly. */

   /*
    * Allow the standard control window procedure to process all
    * other messages.
    */
   default:
      Result = CallWindowProc(lpEditWndProc, hWnd, message, wParam, lParam);

#if FALSE

/*
   If the application superclassed an object that is sensitive to the
   DS that is uses, it is necessary to change the DS at this point.
   CallWindowProc() will change the hInstance for the called procedure
   to that specified in hWnd. When CallWindowProc() returns, DS is
   reset so the application runs with its own DS. Therefore, the
   application can swap the superclassed hInstance allowing the
   superclassed window procedure can use its own DS.

   Modifying the DS occurrs as follows:

   The DispatchMessage() and CallWindowProc() functions look up the
   hInstance from the WND structure and converts it to a DS value placing
   that value in AX. The prologue code for the window procedure executes
   PUSH AX, POP DS, to load DS properly. This means that any window
   procedure for which an instance thunk has not been created with
   MakeProcInstance() will use the DS that corresponds to the hInstance
   of the window structure. The instance thunk inserts a MOV AX,####
   into the prologue code which overrides the value left by
   DispatchMessage() or CallWindowProc().

   MakeProcInstance() should not be used with any application which
   can use multiple instances because the procedure instance will
   be overridded by the application's instance.

   Note: If the application subclasses multiple window procedures in
   multiple instances, it is necessary to verify that the correct
   hInstance is used in all cases.


   To implement this process:

   Before the query phase, when the original window was created to
   retrieve lpEditWndProc, it is also necessary to get
   hEditWndInst through a call to GetWindowWord.

   This information is used at this point in the code as follows:
*/

      /* Swap in the procedure's hInstance/DS. */
      OldhWndInst = SetWindowWord(hWnd, GWW_HINSTANCE, hEditWndInst);
      Result = CallWindowProc(lpEditWndProc, hWnd, message, wParam, lParam);

      /* Restore the hInstance to the application's DS. */
      SetWindowWord(hWnd, GWW_HINSTANCE, OldhWndInst);
#endif
      break;
      }
   return (Result);
}

